@@ -1,30 +1,39 @@
 /****************************************************************************
 **
-** Copyright (C) 2015 The Qt Company Ltd.
-** Contact: http://www.qt.io/licensing
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of Qbs.
 **
+** $QT_BEGIN_LICENSE:LGPL$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms and
-** conditions see http://www.qt.io/terms-conditions. For further information
-** use the contact form at http://www.qt.io/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 or version 3 as published by the Free
-** Software Foundation and appearing in the file LICENSE.LGPLv21 and
-** LICENSE.LGPLv3 included in the packaging of this file.  Please review the
-** following information to ensure the GNU Lesser General Public License
-** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
-** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
-** In addition, as a special exception, The Qt Company gives you certain additional
-** rights.  These rights are described in The Qt Company LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** GNU General Public License Usage
+** Alternatively, this file may be used under the terms of the GNU
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
+**
+** $QT_END_LICENSE$
 **
 ****************************************************************************/
 #include "buildgraphloader.h"
@@ -48,6 +57,7 @@
 #include <logging/translator.h>
 #include <tools/fileinfo.h>
 #include <tools/persistence.h>
+#include <tools/profiling.h>
 #include <tools/propertyfinder.h>
 #include <tools/qbsassert.h>
 
@@ -103,18 +113,27 @@ BuildGraphLoadResult BuildGraphLoader::load(const TopLevelProjectPtr &existingPr
     }
     if (!m_result.loadedProject)
         return m_result;
-    if (parameters.restoreBehavior() == SetupProjectParameters::RestoreOnly)
+    if (parameters.restoreBehavior() == SetupProjectParameters::RestoreOnly) {
+        foreach (const ErrorInfo &e, m_result.loadedProject->warningsEncountered)
+            m_logger.printWarning(e);
         return m_result;
+    }
     QBS_CHECK(parameters.restoreBehavior() == SetupProjectParameters::RestoreAndTrackChanges);
 
+    if (m_parameters.logElapsedTime())
+        m_wildcardExpansionEffort = 0;
     trackProjectChanges();
+    if (m_parameters.logElapsedTime()) {
+        m_logger.qbsLog(LoggerInfo, true) << "\t"
+                << Tr::tr("Wilcard expansion took %1.")
+                   .arg(elapsedTimeString(m_wildcardExpansionEffort));
+    }
     return m_result;
 }
 
 void BuildGraphLoader::loadBuildGraphFromDisk()
 {
-    const QString projectId = TopLevelProject::deriveId(m_parameters.topLevelProfile(),
-                                                        m_parameters.finalBuildConfigurationTree());
+    const QString projectId = TopLevelProject::deriveId(m_parameters.finalBuildConfigurationTree());
     const QString buildDir
             = TopLevelProject::deriveBuildDirectory(m_parameters.buildRoot(), projectId);
     const QString buildGraphFilePath
@@ -199,6 +218,8 @@ static void makeChangedProductsListComplete(QList<ResolvedProductPtr> &changedPr
 
 void BuildGraphLoader::trackProjectChanges()
 {
+    TimedActivityLogger trackingTimer(m_logger, Tr::tr("Change tracking"),
+                                      m_parameters.logElapsedTime());
     const TopLevelProjectPtr &restoredProject = m_result.loadedProject;
     QSet<QString> buildSystemFiles = restoredProject->buildSystemFiles;
     QList<ResolvedProductPtr> allRestoredProducts = restoredProject->allProducts();
@@ -215,7 +236,8 @@ void BuildGraphLoader::trackProjectChanges()
     // can make the list of source files in a product change without the respective file
     // having been touched. In such a case, the build data for that product will have to be set up
     // anew.
-    if (hasBuildSystemFileChanged(buildSystemFiles, restoredProject->lastResolveTime)
+    if (probeExecutionForced(allRestoredProducts)
+            || hasBuildSystemFileChanged(buildSystemFiles, restoredProject->lastResolveTime)
             || hasEnvironmentChanged(restoredProject)
             || hasCanonicalFilePathResultChanged(restoredProject)
             || hasFileExistsResultChanged(restoredProject)
@@ -224,13 +246,20 @@ void BuildGraphLoader::trackProjectChanges()
         reResolvingNecessary = true;
     }
 
-    if (!reResolvingNecessary)
+    if (!reResolvingNecessary) {
+        foreach (const ErrorInfo &e, restoredProject->warningsEncountered)
+            m_logger.printWarning(e);
         return;
+    }
 
     restoredProject->buildData->isDirty = true;
     Loader ldr(m_evalContext->engine(), m_logger);
     ldr.setSearchPaths(m_parameters.searchPaths());
     ldr.setProgressObserver(m_evalContext->observer());
+    QHash<QString, QList<ProbeConstPtr>> restoredProbes;
+    foreach (const auto restoredProduct, allRestoredProducts)
+        restoredProbes.insert(restoredProduct->uniqueName(), restoredProduct->probes);
+    ldr.setOldProbes(restoredProbes);
     m_result.newlyResolvedProject = ldr.loadProject(m_parameters);
 
     QMap<QString, ResolvedProductPtr> freshProductsByName;
@@ -252,8 +281,7 @@ void BuildGraphLoader::trackProjectChanges()
 
             // If the product gets temporarily removed, its artifacts will get disconnected
             // and this structural information will no longer be directly available from them.
-            foreach (const Artifact * const a,
-                     ArtifactSet::fromNodeSet(product->buildData->nodes)) {
+            for (const Artifact *a : filterByType<Artifact>(product->buildData->nodes)) {
                 childLists.insert(a, ChildrenInfo(ArtifactSet::fromNodeSet(a->children),
                                                   a->childrenAddedByScanner));
             }
@@ -320,8 +348,7 @@ void BuildGraphLoader::trackProjectChanges()
 
     foreach (const ResolvedProductConstPtr &changedProduct, changedProducts) {
         rescueOldBuildData(changedProduct, freshProductsByName.value(changedProduct->uniqueName()),
-                           oldBuildData.data(), childLists,
-                           rescuableArtifactData.value(changedProduct->uniqueName()));
+                           childLists, rescuableArtifactData.value(changedProduct->uniqueName()));
     }
 
     EmptyDirectoriesRemover(m_result.newlyResolvedProject.data(), m_logger)
@@ -335,6 +362,17 @@ void BuildGraphLoader::trackProjectChanges()
     doSanityChecks(m_result.newlyResolvedProject, m_logger);
 }
 
+bool BuildGraphLoader::probeExecutionForced(const QList<ResolvedProductPtr> &restoredProducts) const
+{
+    if (!m_parameters.forceProbeExecution())
+        return false;
+    foreach (const auto &p, restoredProducts) {
+        if (!p->probes.isEmpty())
+            return true;
+    }
+    return false;
+}
+
 bool BuildGraphLoader::hasEnvironmentChanged(const TopLevelProjectConstPtr &restoredProject) const
 {
     QProcessEnvironment oldEnv = restoredProject->environment;
@@ -430,6 +468,8 @@ bool BuildGraphLoader::hasProductFileChanged(const QList<ResolvedProductPtr> &re
             m_logger.qbsDebug() << "A product was changed, must re-resolve project";
             hasChanged = true;
         } else if (!changedProducts.contains(product)) {
+            AccumulatingTimer wildcardTimer(m_parameters.logElapsedTime()
+                                            ? &m_wildcardExpansionEffort : nullptr);
             foreach (const GroupPtr &group, product->groups) {
                 if (!group->wildcards)
                     continue;
@@ -521,19 +561,11 @@ bool BuildGraphLoader::checkProductForChanges(const ResolvedProductPtr &restored
     // within commands).
     if (checkForPropertyChanges(restoredProduct, newlyResolvedProduct))
         return true;
-    if (!transformerListsAreEqual(restoredProduct->transformers,
-                                  newlyResolvedProduct->transformers)) {
-        return true;
-    }
     if (!ruleListsAreEqual(restoredProduct->rules.toList(), newlyResolvedProduct->rules.toList()))
         return true;
     if (!dependenciesAreEqual(restoredProduct, newlyResolvedProduct))
         return true;
     const FileTime referenceTime = restoredProduct->topLevelProject()->lastResolveTime;
-    foreach (const ResolvedTransformerConstPtr &transformer, newlyResolvedProduct->transformers) {
-        if (!isPrepareScriptUpToDate(transformer->transform, referenceTime))
-            return true;
-    }
     foreach (const RuleConstPtr &rule, newlyResolvedProduct->rules) {
         if (!isPrepareScriptUpToDate(rule->prepareScript, referenceTime))
             return true;
@@ -591,7 +623,7 @@ bool BuildGraphLoader::checkTransformersForPropertyChanges(const ResolvedProduct
 {
     bool transformerChanges = false;
     QSet<TransformerConstPtr> seenTransformers;
-    foreach (Artifact *artifact, ArtifactSet::fromNodeSet(restoredProduct->buildData->nodes)) {
+    for (Artifact *artifact : filterByType<Artifact>(restoredProduct->buildData->nodes)) {
         const TransformerPtr transformer = artifact->transformer;
         if (!transformer || seenTransformers.contains(transformer))
             continue;
@@ -662,26 +694,45 @@ static QVariantMap propertyMapByKind(const ResolvedProductConstPtr &product, Pro
     return QVariantMap();
 }
 
+static void invalidateTransformer(const TransformerPtr &transformer)
+{
+    const JavaScriptCommandPtr &pseudoCommand = JavaScriptCommand::create();
+    pseudoCommand->setSourceCode(QLatin1String("random stuff that will cause "
+                                               "commandsEqual() to fail"));
+    transformer->commands << pseudoCommand;
+}
+
 bool BuildGraphLoader::checkForPropertyChanges(const TransformerPtr &restoredTrafo,
         const ResolvedProductPtr &freshProduct)
 {
     // This check must come first, as it can prevent build data rescuing.
     foreach (const Property &property, restoredTrafo->propertiesRequestedInCommands) {
         if (checkForPropertyChange(property, propertyMapByKind(freshProduct, property.kind))) {
-            const JavaScriptCommandPtr &pseudoCommand = JavaScriptCommand::create();
-            pseudoCommand->setSourceCode(QLatin1String("random stuff that will cause "
-                                                       "commandsEqual() to fail"));
-            restoredTrafo->commands << pseudoCommand;
+            invalidateTransformer(restoredTrafo);
             return true;
         }
     }
 
+    QMap<QString, SourceArtifactConstPtr> artifactMap;
+    for (auto it = restoredTrafo->propertiesRequestedFromArtifactInCommands.cbegin();
+         it != restoredTrafo->propertiesRequestedFromArtifactInCommands.cend(); ++it) {
+        const SourceArtifactConstPtr artifact
+                = findSourceArtifact(freshProduct, it.key(), artifactMap);
+        if (!artifact)
+            continue;
+        foreach (const Property &property, it.value()) {
+            if (checkForPropertyChange(property, artifact->properties->value())) {
+                invalidateTransformer(restoredTrafo);
+                return true;
+            }
+        }
+    }
+
     foreach (const Property &property, restoredTrafo->propertiesRequestedInPrepareScript) {
         if (checkForPropertyChange(property, propertyMapByKind(freshProduct, property.kind)))
             return true;
     }
 
-    QMap<QString, SourceArtifactConstPtr> artifactMap;
     for (QHash<QString, PropertySet>::ConstIterator it =
          restoredTrafo->propertiesRequestedFromArtifactInPrepareScript.constBegin();
          it != restoredTrafo->propertiesRequestedFromArtifactInPrepareScript.constEnd(); ++it) {
@@ -708,13 +759,8 @@ bool BuildGraphLoader::checkForPropertyChange(const Property &restoredProperty,
         v = newProperties.value(restoredProperty.propertyName);
         break;
     case Property::PropertyInModule:
-        if (restoredProperty.value.type() == QVariant::List) {
-            v = finder.propertyValues(newProperties, restoredProperty.moduleName,
-                                      restoredProperty.propertyName);
-        } else {
-            v = finder.propertyValue(newProperties, restoredProperty.moduleName,
-                                     restoredProperty.propertyName);
-        }
+        v = finder.propertyValue(newProperties, restoredProperty.moduleName,
+                                 restoredProperty.propertyName);
         break;
     }
     if (restoredProperty.value != v) {
@@ -732,14 +778,13 @@ void BuildGraphLoader::replaceFileDependencyWithArtifact(const ResolvedProductPt
 {
     if (m_logger.traceEnabled()) {
         m_logger.qbsTrace()
-            << QString::fromLocal8Bit("[BG] replace file dependency '%1' "
-                                      "with artifact of type '%2'")
-                             .arg(filedep->filePath()).arg(artifact->artifactType);
+            << QString::fromLatin1("[BG] replace file dependency '%1' with artifact of type '%2'")
+               .arg(filedep->filePath()).arg(artifact->artifactType);
     }
     foreach (const ResolvedProductPtr &product, fileDepProduct->topLevelProject()->allProducts()) {
         if (!product->buildData)
             continue;
-        foreach (Artifact *artifactInProduct, ArtifactSet::fromNodeSet(product->buildData->nodes)) {
+        for (Artifact *artifactInProduct : filterByType<Artifact>(product->buildData->nodes)) {
             if (artifactInProduct->fileDependencies.contains(filedep)) {
                 artifactInProduct->fileDependencies.remove(filedep);
                 loggedConnect(artifactInProduct, artifact, m_logger);
@@ -759,10 +804,9 @@ bool BuildGraphLoader::isConfigCompatible()
     for (QVariantMap::ConstIterator it = restoredProject->profileConfigs.constBegin();
          it != restoredProject->profileConfigs.constEnd(); ++it) {
         const QVariantMap buildConfig = SetupProjectParameters::expandedBuildConfiguration(
-                    m_parameters.settingsDirectory(), it.key(), m_parameters.buildVariant());
+                    m_parameters.settingsDirectory(), it.key(), m_parameters.configurationName());
         const QVariantMap newConfig = SetupProjectParameters::finalBuildConfigurationTree(
-                    buildConfig, m_parameters.overriddenValues(), m_parameters.buildRoot(),
-                    m_parameters.topLevelProfile());
+                    buildConfig, m_parameters.overriddenValues(), m_parameters.buildRoot());
         if (newConfig != it.value())
             return false;
     }
@@ -785,79 +829,40 @@ bool BuildGraphLoader::isPrepareScriptUpToDate(const ScriptFunctionConstPtr &scr
 }
 
 void BuildGraphLoader::rescueOldBuildData(const ResolvedProductConstPtr &restoredProduct,
-        const ResolvedProductPtr &newlyResolvedProduct, ProjectBuildData *oldBuildData,
-        const ChildListHash &childLists, const AllRescuableArtifactData &existingRad)
+        const ResolvedProductPtr &newlyResolvedProduct, const ChildListHash &childLists,
+        const AllRescuableArtifactData &existingRad)
 {
     QBS_CHECK(newlyResolvedProduct);
     if (!restoredProduct->enabled || !newlyResolvedProduct->enabled)
         return;
 
     if (m_logger.traceEnabled()) {
-        m_logger.qbsTrace() << QString::fromLocal8Bit("[BG] rescue data of "
-                "product '%1'").arg(restoredProduct->uniqueName());
+        m_logger.qbsTrace() << QString::fromLatin1("[BG] rescue data of product '%1'")
+                               .arg(restoredProduct->uniqueName());
     }
     QBS_CHECK(newlyResolvedProduct->buildData);
     QBS_CHECK(newlyResolvedProduct->buildData->rescuableArtifactData.isEmpty());
     newlyResolvedProduct->buildData->rescuableArtifactData = existingRad;
 
-    // This is needed for artifacts created by manually added transformers, which are
-    // already present in the build graph.
-    // FIXME: This complete block could go away if Transformers were also to create their
-    // output artifacts at build time.
-    QList<Artifact *> oldArtifactsCreatedByTransformers;
-    foreach (Artifact *artifact,
-             ArtifactSet::fromNodeSet(newlyResolvedProduct->buildData->nodes)) {
-        if (!artifact->transformer)
-            continue;
-        if (m_logger.traceEnabled()) {
-            m_logger.qbsTrace() << QString::fromLocal8Bit("[BG]    artifact '%1'")
-                                   .arg(artifact->fileName());
-        }
-
-        Artifact * const oldArtifact = lookupArtifact(restoredProduct, oldBuildData,
-                                                      artifact->dirPath(), artifact->fileName(),
-                                                      true);
-        if (!oldArtifact || !oldArtifact->transformer) {
-            if (m_logger.traceEnabled())
-                m_logger.qbsTrace() << QString::fromLocal8Bit("[BG]    no transformer data");
-            continue;
-        }
-        oldArtifactsCreatedByTransformers << oldArtifact;
-        if (!commandListsAreEqual(artifact->transformer->commands,
-                                  oldArtifact->transformer->commands)) {
-            if (m_logger.traceEnabled())
-                m_logger.qbsTrace() << QString::fromLocal8Bit("[BG]    artifact invalidated");
-            removeGeneratedArtifactFromDisk(oldArtifact, m_logger);
-            m_artifactsRemovedFromDisk << oldArtifact->filePath();
-            continue;
-        }
-        artifact->setTimestamp(oldArtifact->timestamp());
-
-        const ChildrenInfo &childrenInfo = childLists.value(oldArtifact);
-        foreach (Artifact * const oldChild, childrenInfo.children) {
-            Artifact * const newChild = lookupArtifact(oldChild->product,
-                    newlyResolvedProduct->topLevelProject()->buildData.data(),
-                    oldChild->filePath(), true);
-            if (!newChild || artifact->children.contains(newChild))
-                    continue;
-            safeConnect(artifact, newChild, m_logger);
-            if (childrenInfo.childrenAddedByScanner.contains(oldChild))
-                artifact->childrenAddedByScanner << newChild;
-        }
-    }
-
     // This is needed for artifacts created by rules, which happens later in the executor.
-    foreach (Artifact * const oldArtifact,
-             ArtifactSet::fromNodeSet(restoredProduct->buildData->nodes)) {
+    for (Artifact * const oldArtifact : filterByType<Artifact>(restoredProduct->buildData->nodes)) {
         if (!oldArtifact->transformer)
             continue;
-        if (oldArtifactsCreatedByTransformers.contains(oldArtifact))
-            continue;
         Artifact * const newArtifact = lookupArtifact(newlyResolvedProduct, oldArtifact, false);
         if (!newArtifact) {
             RescuableArtifactData rad;
             rad.timeStamp = oldArtifact->timestamp();
+            rad.fileTags = oldArtifact->fileTags();
+            rad.properties = oldArtifact->properties->value();
             rad.commands = oldArtifact->transformer->commands;
+            rad.propertiesRequestedInPrepareScript
+                    = oldArtifact->transformer->propertiesRequestedInPrepareScript;
+            rad.propertiesRequestedInCommands
+                    = oldArtifact->transformer->propertiesRequestedInCommands;
+            rad.propertiesRequestedFromArtifactInPrepareScript
+                    = oldArtifact->transformer->propertiesRequestedFromArtifactInPrepareScript;
+            rad.propertiesRequestedFromArtifactInCommands
+                    = oldArtifact->transformer->propertiesRequestedFromArtifactInCommands;
             const ChildrenInfo &childrenInfo = childLists.value(oldArtifact);
             foreach (Artifact * const child, childrenInfo.children) {
                 rad.children << RescuableArtifactData::ChildData(child->product->name,
